'  ASM WAV Player Ver. 2g  (Plays 16-bit PCM WAV files from SD card)
'  Copyright 2007-2013 Raymond Allen
'  MIT License:   See end of file for terms of use.  
'  Settings for Demo Board Audio Output:  Right Pin# = 10, Left Pin# = 11
'  Rev.2H:  Added support for variable size header before "data"
'  Rev.2C:  Made code easier to use and understand and added mono support 



CON  'Crystal settings
   _clkmode = xtal1 + pll16x
   _xinfreq = 5_000_000       '80 MHz


CON 'SD card pin settings
  'Set the pins used for SD card here
  'Propeller Platform Express settings:
  SD_MISO=0
  SD_CLK=1
  SD_MOSI=2
  SD_CS=3
  
  'Quick Media Player Settings
  {
  SD_MISO=27
  SD_CLK=26
  SD_MOSI=25
  SD_CS=24
  }

CON 'Audio Pins and buffer size
  'Here you can select the audio output pins
  'Demo Board Settings:
  Pin_Right=10
  Pin_Left=11
  'Quick Media Player Settings
  {
  Pin_Right=23
  Pin_Left=22
  } 
  'Slow SD cards may require a larger buffer, adjust with value below
  buffLongs=100
  buffWords = buffLongs*2+1 
  'Volume control
  volume = 65535  'range is 0 to 65535 with 65535 being the loudest   
  

VAR long parameter1  'to pass @buff1 to ASM
    long parameter2  'to pass @buff2 to ASM
    long parameter3  'to pass sample rate to ASM
    long parameter4  'to pass #samples to ASM
    long parameter5  'to pass #channels to ASM
    long parameter6  'to pass #samples in each buffer
    long parameter7  'to pass volume to ASM
    word buff1[buffWords]
    word buff2[buffWords]
    byte Header[44]

    
OBJ
    SD  : "FSRW"
    ser : "fullduplexserial" 'for monitoring

PUB Main
  'Play a WAV File
  'Start serial connection
  ser.start(31,30,0,9600)

  waitcnt(cnt+clkfreq*2)  'Wait 2 seconds (but don't need to)

  ser.str(string(13,13,"Playing wav file...",13))  
  PlayWav(string("test7.wav")) '   <---------  Change .wav filename to play here !!!!!!!!!!!!!!
  ser.str(STRING("Play wav complete.",13))  
  repeat 'all done


PUB PlayWav(pFileName)|n,i,j, SampleRate,Samples,nChannels,ChunkSize,SamplePeriod,PlayTime
  
  'open the WAV file  (NOTE:  Only plays stereo, 16-bit PCM WAV Files !!!!!!!!!!!!!)
  
  'Mounting SD card now
  ser.str(string("Attempting to mount SD card",13))
  repeat   'try forever to mount the SD card
    i:=\sd.Mount_Explicit(SD_MISO,SD_CLK,SD_MOSI,SD_CS) 
    if i==0
       quit  'mounted, so exit
    waitcnt(cnt+clkfreq/4)
    ser.tx(".")

  ser.str(string("SD card mounted.",13)) 



  'open file
  ser.str(string("Attempting to open wav file:  "))
  ser.str(pFileName)
  ser.tx(13)
                   
  i:=\sd.popen(pFileName,"r")             
  if (i<>0)
    ser.str(string("Cannot open wav file.",13))  
    return

  'Plays only   16-bit PCM WAV Files !!!!!!!!!!!!!!!!
  'See here for header format:  https://ccrma.stanford.edu/courses/422/projects/WaveFormat/ 
  i:=sd.pread(@Header, 44-8) 'read data words to input stereo buffer
  'Get bits per sample (must be 16)
  if (Header[35]<<8+Header[34])<>16
    ser.str(string(13,"Error:  Not 16 bits per sample!",13))
    return  
  'Get audio format, must be PCM
  if (Header[21]<<8+Header[20])<>1
    ser.str(string(13,"Error:  Not uncompressed PCM data",13))
    return       
  'Get sample rate from header
  SampleRate:=Header[27]<<24+Header[26]<<16+Header[25]<<8+Header[24]
  ser.str(string("Sample Rate: ")) 
  ser.dec(SampleRate)
  ser.tx(13)
  'Get # channels from header
  nChannels:=Header[23]<<8+Header[22]
  ser.str(string("#Channels: "))
  ser.dec(nChannels)
  ser.tx(13)
  'Find data
  ser.str(string("Looking for data...")) 
  repeat
    sd.pread(@Header[36], 1)
    if Header[36]=="d"
      sd.pread(@Header[37], 1)
      if Header[37]=="a"
        sd.pread(@Header[38], 1)
        if Header[38]=="t"        
          sd.pread(@Header[39], 1)
          if Header[39]=="a"
            quit  'found data
            
  sd.pread(@Header[40],4) 'read subchunk size    
  'get # samples from header
  ChunkSize:=Header[43]<<24+Header[42]<<16+Header[41]<<8+Header[40]
  Samples:=(ChunkSize>>1)/nChannels
  ser.str(string("#Samples: ")) 
  ser.dec(Samples)
  ser.tx(13)
  'Calculate play time
  SamplePeriod:=CLKFREQ/SampleRate  '#clocks between samples'1814'for 44100ksps,  5000 'for 16ksps 
  ser.str(string("Playtime (s) = ")) 
  ser.dec(Samples/SampleRate)
  ser.tx(13)


    

  'Start ASM player in a new cog
  ser.str(STRING("Running ASM Player",13))    
  parameter1:=@buff1                   'address of first buffer
  parameter2:=@buff2                   'address of second buffer
  parameter3:=SamplePeriod             'time between samples
  parameter4:=Samples                  'total # samples
  parameter5:=nChannels                '#channels
  parameter6:=(BuffWords-1)/nChannels  '#samples in each buffer
  parameter7:=volume&$FFFFF            'volume control (0..65535)
  COGNEW(@ASMWAV,@parameter1)

  'Keep filling buffers until end of file
  ' note:  using alternating buffers to keep data always at the ready...
  n:=buffWords-1
  buff1[n]:=1  'set empty flags at end of buffers
  buff2[n]:=1
  i:=n*2   'number of bytes to read

  'Keep buffers filled until got all samples
  j:=samples*nChannels*2
  PlayTime :=cnt
  repeat
    repeat until buff1[n]==1 'wait until buffer empty
    if sd.pread(@buff1, i)<>i  'read data words to input stereo buffer       
      quit 
    buff1[n]:=0 'clear empty flag
    j-=i
    if j=<0
      quit
    
    repeat until buff2[n]==1  'wait until buffer empty
    if sd.pread(@buff2, i)<>i
      quit 'read data words to input stereo buffer   
    buff2[n]:=0 'clear empty flag
    j-=i
    if j=<0
      quit

  ser.str(string("Played time (s) = ")) 
  ser.dec((cnt-PlayTime)/clkfreq)
  ser.tx(13)   
      
  'must have reached the end of the file, so close it
  \sd.pclose

  
  return

 
DAT
  ORG 0
ASMWAV
'load input parameters from hub to cog given address in par
        movd    :par,#pData1             
        mov     x,par
        mov     y,#7  'input 7 parameters
:par    rdlong  0,x
        add     :par,dlsb
        add     x,#4
        djnz    y,#:par

setup
        'setup output pins
        MOV DMaskR,#1
        ROL DMaskR,OPinR
        OR DIRA, DMaskR
        MOV DMaskL,#1
        ROL DMaskL,OPinL
        OR DIRA, DMaskL
        'setup counters
        OR CountModeR,OPinR
        MOV CTRA,CountModeR
        OR CountModeL,OPinL
        MOV CTRB,CountModeL
        'Wait for SPIN to fill table
        MOV WaitCount, CNT
        ADD WaitCount,BigWait
        WAITCNT WaitCount,#0
        'setup loop table
        MOV LoopCount,SizeBuff
        MOV pData,pData1
        MOV nTable,#1
        'setup loop counter
        MOV WaitCount, CNT
        ADD WaitCount,dRate



MainLoop
        CMP nSamples,#0 wz,wc 'see if out of samples
        IF_e JMP #Done
        SUB nSamples,#1
        waitcnt WaitCount,dRate
        cmp   nChannel,#2 wz,wc
        if_e jmp #Stereo     'otherwise, assume mono
Mono
        call  #ReadWithVolume
        mov   right,y
        mov   left,right
        jmp #Output
        
Stereo
        call  #ReadWithVolume
        mov   right,y
        add pData,#2
        call  #ReadWithVolume
        mov   left,y 

Output
        ADD pData,#2       
        MOV FRQA,Right        
        MOV FRQB,Left
        

        'loop
        DJNZ LoopCount,#MainLoop

        WRword one,pData  'set empty flag 
        
        MOV LoopCount,SizeBuff        
        'switch table       ?
        CMP nTable,#1 wz
        IF_Z JMP #SwitchToTable2
SwitchToTable1
        MOV nTable,#1
        MOV pData,pData1
        JMP #MainLoop
SwitchToTable2
        MOV nTable,#2
        MOV pData,pData2
        JMP #MainLoop
        
                
Done
        
         'now stop
        COGID thisCog
        COGSTOP thisCog          

'Multiply routine adapted from manual
'' Multiply x[15..0] by y[15..0] (y[31..16] must be 0)
' on exit, product in y[31..0]
ReadWithVolume
        mov   x,dVolume
        rdword y,pData
        add   y,twos
        and   y,ffff
        shl x,#16 'get multiplicand into x[31..16]
        mov t,#16 'ready for 8 multiplier bits
        shr y,#1 wc 'get initial multiplier bit into c
:loop
        if_c add y,x wc 'if c set, add multiplicand to product
        rcr y,#1 wc 'put next multiplier in c, shift prod.
        djnz t,#:loop 'loop until done

ReadWithVolume_ret
        ret 'return with product in y[31..0]


'Working variables
thisCog long 0
x       long 0
y       long 0
t       long 0
dlsb    long    1 << 9
BigWait long 100000
twos    long $8000_8000
ffff    long $ffff
        
'Loop parameters
nTable  long 0
WaitCount long 0
pData   long 0
LoopCount long 0
Left    long 0
Right   long 0
Zero    long 0
one     long 1       

'setup parameters
DMaskR  long 0 'right output mask
OPinR   long Pin_Right 'right channel output pin #                   
DMaskL  long 0 'left output mask 
OPinL   long Pin_Left 'left channel output pin #                       
CountModeR long %00011000_00000000_00000000_00000000
CountModeL long %00011000_00000000_00000000_00000000


'input parameters
pData1   res 1  'Address of first data table        
pData2   res 1 'Address of second data table
dRate    res 1 'clocks between samples
nSamples res 1 '#samples to play
nChannel res 1 '#channels (1 or 2)
SizeBuff res 1 '#samples in each buffer
dVolume  res 1 'volume byte



DAT 'MIT License
{{
'  Permission is hereby granted, free of charge, to any person obtaining
'  a copy of this software and associated documentation files
'  (the "Software"), to deal in the Software without restriction,
'  including without limitation the rights to use, copy, modify, merge,
'  publish, distribute, sublicense, and/or sell copies of the Software,
'  and to permit persons to whom the Software is furnished to do so,
'  subject to the following conditions:
'
'  The above copyright notice and this permission notice shall be included
'  in all copies or substantial portions of the Software.
'
'  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
'  EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
'  MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
'  IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY
'  CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT,
'  TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE
'  SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
}}